/*
    Copyright 2010 Luigi Auriemma

    This program is free software; you can redistribute it and/or modify
    it under the terms of the GNU General Public License as published by
    the Free Software Foundation; either version 2 of the License, or
    (at your option) any later version.

    This program is distributed in the hope that it will be useful,
    but WITHOUT ANY WARRANTY; without even the implied warranty of
    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
    GNU General Public License for more details.

    You should have received a copy of the GNU General Public License
    along with this program; if not, write to the Free Software
    Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307 USA

    http://www.gnu.org/licenses/gpl-2.0.txt
*/

#include <stdio.h>
#include <stdlib.h>
#include <string.h>
#include <stdint.h>
#include <sys/stat.h>
#include "show_dump.h"

#ifdef WIN32
    #include <winsock.h>
    #include "winerr.h"

    #define close   closesocket
    #define sleep   Sleep
    #define ONESEC  1000
    #define sleepms sleep
#else
    #include <unistd.h>
    #include <sys/socket.h>
    #include <sys/types.h>
    #include <arpa/inet.h>
    #include <netinet/in.h>
    #include <netdb.h>

    #define stricmp strcasecmp
    #define stristr strcasestr
    #define ONESEC  1
    #define sleepms(X)  usleep(X * 1000)
#endif

typedef uint8_t     u8;
typedef uint16_t    u16;
typedef uint32_t    u32;



#define VER         "0.2a"
#define BUFFSZ      0xffff
#define PORT        69
#define TIMEOUT     2
#define NONE        "-"
#define GETXX16(X)  ((buff[X] << 8) | buff[X + 1])
#define SENDTO(X)   if(hexdump & 2) show_dump(buff, X, stdout); \
                    if(sendto(sd, buff, X, 0, (struct sockaddr *)peer, sizeof(struct sockaddr_in)) \
                      < 0) std_err();



u32 tftpx(FILE *fd, int sd, struct sockaddr_in *peer, int upload, u8 *buff, u32 filesz, int blksize);
u8 *bof_string(u8 *str, int chr, int *ret_size);
u32 get_num(u8 *str);
int putss(u8 *data, u8 *str);
int putcc(u8 *data, int chr, int size);
int putxx(u8 *data, u32 num, int bits);
int timeout(int sock);
u32 resolv(char *host);
void std_err(void);



int     hexdump = 0;



int main(int argc, char *argv[]) {
    struct  sockaddr_in *peer;
    struct  stat    xstat;
    FILE    *fd         = NULL;
    u32     tsize       = 0;
    int     sd,
            i,
            upload      = 0,
            overwrite   = 0,
            bofsize     = 0,
            opcode      = 0xffff + 1,   // do NOT touch int and value!
            n_options   = 0,
            fixedsize   = -1,
            chr         = 'a',
            loop        = -1;
    u16     port        = PORT;
    u8      ans[16],
            *buff,
            *host,
            *local,
            *remote,
            *blksize    = NULL,
            *tout       = NULL,
            *multicast  = NULL,
            *transfer   = "octet",  // octet, netasci, mail
            *myoption   = NULL,
            *myvalue    = NULL,
            *p;

#ifdef WIN32
    WSADATA    wsadata;
    WSAStartup(MAKEWORD(1,0), &wsadata);
#endif

    setbuf(stderr, NULL);
    setbuf(stdout, NULL);
    setbuf(stdin,  NULL);

    fputs("\n"
        "TFTP server tester "VER"\n"
        "by Luigi Auriemma\n"
        "e-mail: aluigi@autistici.org\n"
        "web:    aluigi.org\n"
        "\n", stderr);

    if(argc < 4) {
        fprintf(stderr, "\n"
            "Usage: %s [options] <host> <remote_file> <local_file>\n"
            "\n"
            "-u        upload a file (opcode 2), default is download (opcode 1)\n"
            "-t SIZE   \"tsize\" option, default is %u or real size if upload\n"
            "-b SIZE   \"blksize\" option, default is not set\n"
            "-o NUM    \"timeout\" option, default is not set\n"
            "-m NUM    \"multicast\" option, default is not set\n"
            "-c X Y    add a custom value where X is the option and Y its value\n"
            "-C X Y    like above but X and Y are the size of the 2 values filled with '%c'\n"
            "-p PORT   server port, default is %hu, you can specify it also in host:port\n"
            "-x        show the hexdump of any packet received\n"
            "-X        show the hexdump of any packet sent\n"
            "-y        automatically overwrite the local file if exists (only download)\n"
            "-f        this option is useful to easily test possible buffer-overflows in the\n"
            "          filename sent to the server without manually specifying it. The\n"
            "          default char is '%c' (0x%02x) and the number of chars to compose the\n"
            "          filename must be specified in remote_file, eg: -f SERVER 1000 myfile\n"
            "-C CHR    change the default char used for the strings built by their size\n"
            "-O NUM    custom 16 bit opcode number\n"
            "-T STR    custom tranfer type like \"octet\", \"netascii\", \"mail\" and so on\n"
            "          if you use a number it will be handled as a string of STR chars\n"
            "-N NUM    add NUM fake options in the packet\n"
            "-S SIZE   set a fixed size that must have the packet, so doesn't matter what\n"
            "          is the size of the resulted packet, it will be sent with SIZE bytes\n"
            "-L MS     loop the initial request using a delay of MS milliseconds\n"
            "\n"
            " use "NONE" as local_file for using stdout (download) and stdin (upload)\n"
            " the tool accepts both decimal and hexadecimal numbers\n"
            "\n", argv[0], tsize, chr, port, chr, chr);
        exit(1);
    }

    argc -= 3;
    for(i = 1; i < argc; i++) {
        if(((argv[i][0] != '-') && (argv[i][0] != '/')) || (strlen(argv[i]) != 2)) {
            fprintf(stderr, "\nError: wrong argument (%s)\n", argv[i]);
            exit(1);
        }
        switch(argv[i][1]) {
            case '-':
            case '?':
            case 'h': {
                fprintf(stderr, "\nError: use no arguments for the runtime help\n");
                exit(1);
                break;
            }
            case 'u': upload    = 1;                    break;
            case 't': tsize     = get_num(argv[++i]);   break;
            case 'b': blksize   = argv[++i];            break;
            case 'o': tout      = argv[++i];            break;
            case 'm': multicast = argv[++i];            break;
            case 'c': {
                myoption        = argv[++i];
                myvalue         = argv[++i];
                break;
            }
            case 'C': {
                myoption        = bof_string(argv[++i], chr, NULL);
                myvalue         = bof_string(argv[++i], chr, NULL);
                break;
            }
            case 'p': port      = get_num(argv[++i]);   break;
            case 'x': hexdump   |= 1;                   break;
            case 'X': hexdump   |= 2;                   break;
            case 'y': overwrite = 1;                    break;
            case 'f': bofsize   = 1;                    break;
            case 'O': opcode    = get_num(argv[++i]);   break;
            case 'T': transfer  = argv[++i];            break;
            case 'N': n_options = get_num(argv[++i]);   break;
            case 'S': fixedsize = get_num(argv[++i]);   break;
            case 'L': loop      = get_num(argv[++i]);   break;
            default: {
                fprintf(stderr, "\nError: Wrong command-line argument (%s)\n\n", argv[i]);
                exit(1);
            } break;
        }
    }

    host   = argv[argc];
    remote = argv[argc + 1];
    local  = argv[argc + 2];

    p = strchr(host, ':');
    if(p) {
        port = atoi(p + 1);
        *p = 0;
    }
    peer = calloc(sizeof(struct sockaddr_in), 1);   // good for SEND macro
    peer->sin_addr.s_addr = resolv(host);
    peer->sin_port        = htons(port);
    peer->sin_family      = AF_INET;

    fprintf(stderr, "- target             %s : %u\n",
        inet_ntoa(peer->sin_addr), ntohs(peer->sin_port));

    if(bofsize) {
        remote = bof_string(remote, chr, &bofsize);
        fprintf(stderr, "- size of filename:  %d\n", bofsize);
    }
    fprintf(stderr, "- remote file:       %s\n", remote);

    if(!strcmp(local, NONE)) {
        if(!upload) {
            fprintf(stderr, "- local file:        stdout\n");
            fd = stdout;
        } else {
            fprintf(stderr, "- local file:        stdin\n");
            fd = stdin;
        }
    } else {
        fprintf(stderr, "- local file:        %s\n", local);
        if(!upload) {
            if(!overwrite) {
                fd = fopen(local, "rb");
                if(fd) {
                    fclose(fd);
                    fprintf(stderr, "- file exists, do you want to overwrite it?\n  (y/N) ");
                    fgets(ans, sizeof(ans), stdin);
                    if((ans[0] != 'y') && (ans[0] != 'Y')) {
                        fprintf(stderr, "- exit\n\n");
                        exit(1);
                    }
                }
            }
            fd = fopen(local, "wb");
        } else {
            fd = fopen(local, "rb");
        }
        if(!fd) std_err();
    }

    if(opcode <= 0xffff) {  // yes, you can use also -1 for 0xffff
        // already set
    } else if(!upload) {
        opcode = 1; // RRQ
    } else {
        opcode = 2; // WRQ
        if(!tsize) {
            if(fd == stdin) {
                tsize = -1;
            } else {
                fstat(fileno(fd), &xstat);
                tsize = xstat.st_size;
            }
            fprintf(stderr, "- local file size:   %u\n", tsize);
        }
    }

    if((transfer[0] >= '0') && (transfer[0] <= '9')) {
        transfer = bof_string(transfer, chr, NULL);
    }

    buff = malloc(BUFFSZ);
    if(!buff) std_err();

    sd = socket(AF_INET, SOCK_DGRAM, IPPROTO_UDP);
    if(sd < 0) std_err();

    p = buff;
    p += putxx(p, opcode, 16);
    p += putss(p, remote);  // it's good to have it here even if not RRQ/WRQ opcode so that
                            // during the tests with other opcodes is possible to use it

    if((opcode == 1) || (opcode == 2)) {    // normal download/upload
        p += putss(p, transfer);

        p += putss(p, "tsize");
        p += sprintf(p, "%u", tsize) + 1;

        if(blksize) {
            p += putss(p, "blksize");
            p += putss(p, blksize);
        }
        if(tout) {
            p += putss(p, "timeout");
            p += putss(p, tout);
        }
        if(multicast) {
            p += putss(p, "multicast");
            p += putss(p, multicast);
        }
        if(myoption) {
            p += putss(p, myoption);
            p += putss(p, myvalue);
        }
        if(n_options) {
            for(i = 0; i < n_options; i++) {
                if(((p + 4) - buff) >= BUFFSZ) break;   // useful?
                p += putss(p, "x");
                p += putss(p, "y");
            }
        }
    }

    if((p - buff) > BUFFSZ) {
        fprintf(stderr, "\nError: the built packet (%u) is bigger than UDP packet (%u)\n", p - buff, BUFFSZ);
        exit(1);
    }
    if(fixedsize >= 0) {
        if(fixedsize > BUFFSZ) fixedsize = BUFFSZ;
        i = p - buff;
        if(i < fixedsize) {
            p += putcc(p, chr, fixedsize - i);
        } else {
            p = buff + fixedsize;
        }
    }

    fprintf(stderr, "- send file request (%d bytes)\n", p - buff);
    if(loop >= 0) {
        for(;;) {
            SENDTO(p - buff)
            sleepms(loop);
        }
        return(0);  // it will never terminate
    }
    for(i = 2; i > 0; i--) {
        SENDTO(p - buff)
        if(timeout(sd) >= 0) break;
    }
    if(i <= 0) {
        fprintf(stderr, "\nError: socket timeout, no reply received from the server\n");
        exit(1);
    }

    fprintf(stderr, "- start the file transfer:\n");
    tsize = tftpx(fd, sd, peer, upload, buff, tsize, blksize ? get_num(blksize) : -1);
    fprintf(stderr, "\n- %u bytes %s\n", tsize, upload ? "sent" : "received");

    if(fd && (fd != stdin) && (fd != stdout)) fclose(fd);
    close(sd);
    return(0);
}



u32 tftpx(FILE *fd, int sd, struct sockaddr_in *peer, int upload, u8 *buff, u32 filesz, int blksize) {
    u32     tot;
    int     len,
            psz,
            num;
    u16     opcode,
            block;

    if(blksize < 0) blksize = 512;  // default
    if(blksize > (BUFFSZ - 4)) blksize = BUFFSZ - 4;
    num = 0;
    tot = 0;
    if(!upload) filesz  = -1;
    while(tot < filesz) {
        if(timeout(sd) < 0) {
            fprintf(stderr, "\n- timeout or download finished\n");
            break;
        }
        psz = sizeof(struct sockaddr_in);
        len = recvfrom(sd, buff, BUFFSZ, 0, (struct sockaddr *)peer, &psz);
        if(len < 0) std_err();
        if(!len) break;
        if(hexdump & 1) show_dump(buff, len, stdout);
        if(len < 4) continue;

        opcode = GETXX16(0);
        block  = GETXX16(2);
        if(!upload) {
            switch(opcode) {
                case 3: {       // DATA
                    len -= 4;
                    if(fwrite(buff + 4, 1, len, fd) != len) {
                        fprintf(stderr, "\nError: impossible to write into the local file\n");
                        exit(1);
                    }
                    fflush(fd);
                    tot += len;
                    break;
                }
                case 5: {       // ERROR
                    fprintf(stderr, "\n"
                        "Error: TFTP error %d from the server:\n"
                        "\n"
                        "  %.*s\n"
                        "\n", block, len - 4, buff + 4);
                    exit(1);
                    break;
                }
                case 6: {       // OACK
                    if(!stricmp(buff + 2, "tsize")) {
                        filesz = get_num(buff + 8);
                        fprintf(stderr, "\n- remote file size:  %u\n", filesz);
                    }
                    putxx(buff + 2, 0, 16);
                    break;
                }
                default: {
                    fprintf(stderr, "\n- unknown tftp opcode (%u)\n\n", opcode);
                    if(!(hexdump & 1)) show_dump(buff, len, stdout);
                    //exit(1);
                    break;
                }
            }

            putxx(buff, 4, 16);
            len = 0;
        } else {
            switch(opcode) {
                case 4: {       // ACK
                    if(block != num) {
                        //fprintf(stderr, "\nError: packet lost, retransmission is not supported yet\n");
                        //exit(1);
                        // the following is only a simple and probably wrong tentative
                        num = block;
                        fprintf(stderr, "\n- packet %d lost, I try to resend from that point\n", num);
                        if(fd != stdin) {
                            tot = num * blksize;
                            if(fseek(fd, tot, SEEK_SET)) std_err();
                        }
                    }
                    break;
                }
                case 5: {       // ERROR
                    fprintf(stderr, "\n"
                        "Error: TFTP error %d from the server:\n"
                        "\n"
                        "  %.*s\n"
                        "\n", block, len - 4, buff + 4);
                    exit(1);
                    break;
                }
                case 6: {       // OACK
                    break;
                }
                default: {
                    fprintf(stderr, "\n- unknown tftp opcode (%u)\n\n", opcode);
                    if(!(hexdump & 1)) show_dump(buff, len, stdout);
                    //exit(1);
                    break;
                }
            }

            len = fread(buff + 4, 1, blksize, fd);
            if(len <= 0) break;
            tot += len;

            putxx(buff, 3, 16);
            putxx(buff + 2, ++num, 16);
        }

        fputc('.', stderr);
        SENDTO(4 + len)
    }

    if(upload /*&& (tot == filesz)*/) { // seems needed or the server doesn't close the file
        putxx(buff, 3, 16);
        putxx(buff, ++num, 16);
        SENDTO(4)
    }
    return(tot);
}



u8 *bof_string(u8 *str, int chr, int *ret_size) {
    int     len;
    u8      *ret;

    len = get_num(str);
    if((len < 0) || (len > BUFFSZ)) len = BUFFSZ;   // or return NULL?
    ret = malloc(len + 1);
    if(!ret) std_err();
    memset(ret, chr, len);
    ret[len] = 0;
    if(ret_size) *ret_size = len;
    return(ret);
}



u32 get_num(u8 *str) {
    u32     ret;
    u8      *p; // in case of old libc

    p = strchr(str, 'x');
    if(p) {
        sscanf(p + 1, "%x", &ret);
    } else {
        ret = atoi(str);
    }
    return(ret);
}



int putss(u8 *data, u8 *str) {
    int     len;

    len = strlen(str) + 1;
    memcpy(data, str, len);
    return(len);
}



int putcc(u8 *data, int chr, int size) {
    memset(data, chr, size);
    return(size);
}



int putxx(u8 *data, u32 num, int bits) {
    int     i,
            bytes;

    bytes = bits >> 3;
    for(i = 0; i < bytes; i++) {
        data[i] = num >> ((bytes - 1 - i) << 3);    // big
        //data[i] = num >> (i << 3);    // little
    }
    return(bytes);
}



int timeout(int sock) {
    struct  timeval tout;
    fd_set  fd_read;
    int     err;

    tout.tv_sec = TIMEOUT;
    tout.tv_usec = 0;
    FD_ZERO(&fd_read);
    FD_SET(sock, &fd_read);
    err = select(sock + 1, &fd_read, NULL, NULL, &tout);
    if(err < 0) std_err();
    if(!err) return(-1);
    return(0);
}



u32 resolv(char *host) {
    struct hostent *hp;
    u32 host_ip;

    host_ip = inet_addr(host);
    if(host_ip == INADDR_NONE) {
        hp = gethostbyname(host);
        if(!hp) {
            fprintf(stderr, "\nError: Unable to resolv hostname (%s)\n", host);
            exit(1);
        } else host_ip = *(u32 *)hp->h_addr;
    }
    return(host_ip);
}



#ifndef WIN32
    void std_err(void) {
        perror("\nError");
        exit(1);
    }
#endif


